f1 <- function(t) {5*t}
f2 <- function(t) {3*(2*t-1)^2}
f3 <- function(t) {4*sin(2*pi*t)/(2-sin(2*pi*t))}
f4 <- function(t) {6*(0.1*sin(2*pi*t)+0.2*cos(2*pi*t)+0.3*(sin(2*pi*t))^2+
                        0.4*(cos(2*pi*t))^3+0.5*(sin(2*pi*t))^3)}

GFI_samples <- function(X, y, q=5/ncol(X), l=3, K=6) {
  # p_max: the maximum size of a model candidate
  n <- nrow(X); p <- ncol(X)
  m_n <- K+l
  p_max <- floor(n/m_n)
  bs_X <- matrix( rep(0, n*m_n*p), nrow=n)
  for (i in 1:p) {
    bs_X[,(m_n*i-m_n+1):(m_n*i)] <- bs(X[,i], df = m_n, degree = l)
  }
  bs_X <- scale(bs_X, scale=F)
  
  generate_models <- function(bs_X, y){
    gl <- gglasso(bs_X, y, group=rep(1:p, each = m_n),loss="ls",intercept=T)
    models <- apply(gl$beta, 2, function(b){
      sapply(1:p, function(i) sum(abs(b[((i-1)*m_n+1):(i*m_n)])) !=0)
    })
    unique_models <- unique(data.frame(t(models)))
  }
  
  cat("generate unique models...\n")
  models <- foreach(i=1:10, .combine=rbind) %do%{
    row <- sample(1:n, n, replace=TRUE)
    generate_models(bs_X[row,], y[row])
  }
  unique_models <- unique(models)
  sizes <- apply(unique_models, 1, sum)
  
  gfd <- function(index){
    X1 <- do.call(cbind, lapply(index, function(i) bs_X[,(m_n*i-m_n+1):(m_n*i)]))
    ls_fit <- lm(y~X1)
    RSS <- sum((y-ls_fit$fitted)^2)
    p_star <- m_n*length(index)
    ((p_star-n)/2)*log(2*pi)+(n-p_star-2)/2*log(2)+(p_star-n+1)/2*log(RSS)+lgamma((n-p_star)/2)+p_star*log(q)
  }
  
  cat("compute fiducial densities...\n")
  r <- sapply(1:nrow(unique_models), function(j){
    index <- which(as.logical(unique_models[j,]))
    if (length(index)==0 || length(index)>p_max){
      return(NA)
    }
    gfd(index)
  })
  
  pr <- exp(r-max(r, na.rm=T))
  pr <- pr/sum(pr, na.rm=T)
  
  selected_f <- which(as.logical(unique_models[which.max(pr),]))
  cat("The best model is ", selected_f, "\n")
  
  model_cache = list()
  get_model <- function(i){
    if (i > length(model_cache) || is.null(model_cache[[i]])){
      index <- which(unique_models[i,]!=0)
      X1 <- do.call(cbind, lapply(index, function(i) { bs_X[,(m_n*i-m_n+1):(m_n*i)] }))
      ls_fit <- lm(y~X1)
      RSS <- sum((y-fitted(ls_fit))^2)
      p_star <- m_n*length(index)
      b_est <- coef(ls_fit)[-1]
      model_cache[[i]] <<- list(index=index, X1=X1, RSS=RSS, p_star=p_star, b_est=b_est)
    }
    model_cache[[i]]
  }
  
  cat("generate fiducial samples...\n")
  generate_samples <- function(){
    index_sample=sample(1:length(pr), 1, replace=T, prob=pr)
    model <- get_model(index_sample)
    index <- model$index
    X1 <- model$X1
    RSS <- model$RSS
    p_star <- model$p_star
    var_sample <- RSS/rchisq(n=1, df=n-p_star)
    .beta_sample <- mvrnorm(n = 1, model$b_est, solve(t(X1)%*%X1)*var_sample)
    beta_sample <- rep(0, p*m_n)
    for (i in seq_along(index)){
      ind <- index[i]
      beta_sample[(m_n*ind-m_n+1):(m_n*ind)] <- .beta_sample[(m_n*i-m_n+1):(m_n*i)]
    }
    # y_est <- X1 %*% beta_sample
    # see if the best model is selected
    #best_ind <- setequal(c(1,2,3,4), index)
    return(list(model_index=index, var_sample=var_sample, beta_sample=beta_sample))#, best_ind=best_ind))
  }
  
  pr[is.na(pr)]<-0
  samples <- foreach(i=1:10000) %do%{
    generate_samples()
  }
  
  return(samples)
}

sim_coverage <- function(sd=1, alpha=0.05, l=3, K=6) {
  m_n=l+K
  n <- nrow(X); p <- ncol(X)
  y <- Ey+rnorm(n, sd=sd)
  samples <- GFI_samples(X, y, l=l, K=K)
  
  model_ind <- unlist(lapply(samples, function(x) {x[[1]]}))
  f1_prob <- sum(model_ind==1)/10000
  f2_prob <- sum(model_ind==2)/10000
  f3_prob <- sum(model_ind==3)/10000
  f4_prob <- sum(model_ind==4)/10000
  
  var_samples <- sapply(samples, function(x) {x[[2]]})
  var_bounds <- quantile(var_samples, c(alpha/2 ,1-alpha/2))
  var_bias <- var_samples-sd^2 
  var_bias_avg <- mean(var_bias)
  var_bias_sd <- sd(var_bias)
  var_width <- var_bounds[2] - var_bounds[1]
  var_ind <- as.numeric((var_bounds[1]<sd^2) & (var_bounds[2]>sd^2))
  
  beta_samples <- sapply(samples, function(x) {x[[3]]})
  bs_X <- matrix( rep(0, n*m_n*p), nrow=n)
  for (j in 1:p) {
    bs_X[,(m_n*j-m_n+1):(m_n*j)] <- bs(X[,j], df = m_n, degree = l)
  }
  bs_X <- scale(bs_X, scale=F)
  Ey_samples <- bs_X%*%beta_samples + mean(y)
  Ey_conf <- t(apply(Ey_samples, 1, probs=c(alpha/2 ,1-alpha/2), quantile))
  Ey_ind <- sapply(1:n, function(i) as.numeric((Ey_conf[i,1]<Ey[i]) & (Ey_conf[i,2]>Ey[i])))
  Ey_width <- Ey_conf[,2] - Ey_conf[,1]
  
  #using classic theory
  fit0 <- lm(y~bs_X[, 1:(4*m_n)])
  sigma0 <-summary(fit0)$sigma 
  b_est <- coef(fit0)[-1]
  Ey0 <- bs_X[, 1:(4*m_n)]%*%b_est + mean(y)
  X0 <- bs_X[, 1:(4*m_n)]
  se0 <- sigma0*sqrt(diag(X0%*%solve(t(X0)%*%X0)%*%t(X0)))
  qt0<-qnorm(1-alpha/2)
  Ey0_conf <- cbind(Ey0-qt0*se0, Ey0+qt0*se0)
  Ey0_ind <- sapply(1:n, function(i) as.numeric((Ey0_conf[i,1]<Ey[i]) & (Ey0_conf[i,2]>Ey[i])))
  Ey0_width <- 2*qt0*se0
  
  #confidence interval for variance
  var0_bounds <- c((n-(4*m_n))*sigma0^2/qchisq(1-alpha/2,n-4*m_n), (n-4*m_n)*sigma0^2/qchisq(alpha/2,n-4*m_n))
  var0_width <- var0_bounds[2]-var0_bounds[1]
  var0_ind <- as.numeric((var0_bounds[1]<sd^2) & (var0_bounds[2]>sd^2))
  
  f_ind <- matrix(rep(0, n*4), nrow=4)
  f_width <- matrix(rep(0, n*4), nrow=4)
  for (i in 1:4) {
    betai <- beta_samples[(m_n*i-m_n+1):(m_n*i),]
    Xi <- sort(X[,i])
    BXi <- bs(Xi, df = m_n, degree = l)
    BXi <- scale(BXi, scale=F)
    bounds <- sapply(1:n, function(j){
      quantile(BXi[j,]%*%betai, c(alpha/2 ,1-alpha/2))
    })
    
    fi <- get(paste("f",i,sep=""))
    f_ind[i,] <- (bounds[1,]+mean(fi(Xi)) < fi(Xi)) & (fi(Xi) < bounds[2,]+mean(fi(Xi)))
    f_width[i,] <- bounds[2,]-bounds[1,]
  }  
  
  return(list( f_ind=f_ind, f_width=f_width, sigma0=sigma0,
            var_ind=var_ind, var_width=var_width, var0_ind=var0_ind, var0_width=var0_width,
            Ey_ind=Ey_ind, Ey_width=Ey_width,Ey0_ind=Ey0_ind, Ey0_width=Ey0_width,
            f1_prob=f1_prob, f2_prob=f2_prob, f3_prob=f3_prob, f4_prob=f4_prob))
}

plot_ci <- function(X, beta_samples, i=4, l=3, K=8){
  m_n=l+K
  betai <- beta_samples[(m_n*i-m_n+1):(m_n*i),]
  Xi <- sort(X[,i])
  BXi <- bs(Xi, df = m_n, degree = l)
  BXi <- scale(BXi, scale=F)
  bounds <- sapply(1:n, function(j){
    quantile(BXi[j,]%*%betai, c(0.025 ,0.975))
  })
  fi <- get(paste("f",i,sep=""))
  curve(fi, ylim=range(bounds)+mean(fi(Xi)), ylab=expression("f"[5]*"(x)"), mgp=c(1.8,0.5,0))
  lines(Xi, bounds[1,]+mean(fi(Xi)), col="red")
  lines(Xi, bounds[2,]+mean(fi(Xi)), col="red")
}